#include <stdlib.h>
#include <stdio.h>
#include <sys/shm.h>
#include <unistd.h>
#include "shared.h"

void StartConsumer(SharedStruct* sharedMem);
void Consume(SharedStruct* sharedMem);
int GetNextUsedSlot(SharedStruct* sharedMem);
void SemaphoreCleanup(SharedStruct* sharedMem);
void consume_item(int item, SharedStruct* sharedMem);
int remove_item(SharedStruct* sharedMem);

// Checks for SharedID, then begins producing numbers into shared memory.
int main(int argc, char* argv[])
{
    if (argc != 2)
    {
        fprintf(stderr, "Usage: %s SharedID# \n", argv[0]);
        exit(1);
    }
    int sharedID = StringToNumber(argv[1]);
    SharedStruct* sharedMem = shmat(sharedID, NULL, 0);
    StartConsumer(sharedMem);
    PrintNumberCount(sharedMem);
    return 0;
}

//  Starts the consumer process.
void StartConsumer(SharedStruct* sharedMem)
{
    while (!IsProductionFinished(sharedMem))
    {
        sem_wait(&sharedMem->full);
        Consume(sharedMem);
        sem_post(&sharedMem->empty);
    }
    SemaphoreCleanup(sharedMem);
}

// Takes a number from the shared memory and consumes it.
void Consume(SharedStruct* sharedMem)
{
    int consumeNum;
    sem_wait(&sharedMem->mutex);
    consumeNum = remove_item(sharedMem);
    if (consumeNum >= 0)
        consume_item(consumeNum, sharedMem);
    sem_post(&sharedMem->mutex);
    return;
}

int GetNextUsedSlot(SharedStruct* sharedMem)
{
    int slot = -1;
    for (int i = 0; i < 10; i++)
    {
        if (sharedMem->buffer[i] > -1)
        {
            slot = i;
            return slot;
        }
    }
    return slot;
}

// Weird semaphore math that cleans up the stuck forks
void SemaphoreCleanup(SharedStruct* sharedMem)
{
    int semMath = 2*(sharedMem->consumerCount) + (sharedMem->producerCount) - 1;
    int semCheckEmpty;
    sem_getvalue(&sharedMem->empty, &semCheckEmpty);
    if (semCheckEmpty < semMath)
    {
        sem_post(&sharedMem->full);
        exit(0);
    }
    return;
}

//  Increments the count of numbers generated so far in shared memory.
void consume_item(int item, SharedStruct* sharedMem)
{
    sharedMem->count[item]++;
    return;
}

//  Take item out of shared memory.
int remove_item(SharedStruct* sharedMem)
{
    int slot = GetNextUsedSlot(sharedMem);
    if (slot < 0)
        return -1;
    int item = sharedMem->buffer[slot];
    sharedMem->buffer[slot] = -1;
    return item;
}